# -*- coding: utf-8 -*-

from chef import autoconfigure, Search
from datetime import datetime
from fabric import api
from mls.fabfile.exceptions import err


def mls_config():
    """Get the MLS user for the current node."""
    chef_api = autoconfigure()
    if '127.0.0.1:2222' in api.env.hosts:
        query = 'hostname:%s' % api.env.hostname
    else:
        query = 'ipaddress:%s' % api.env.host
    for node in Search('node', query, api=chef_api):
        try:
            return node.object['mls']
        except KeyError:
            return {}


def backup_dev_packages(folder=None, user=None):
    """Backup the development packages."""
    config = mls_config()
    folder = folder or config.get('app', {}).get('dir') or err('Folder must be set!')
    user = user or config.get('user') or err('MLS user must be set!')
    with api.settings(sudo_user=user):
        # Backup buildout src packages.
        with api.cd(folder):
            now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            backup_folder = 'src_backups/%s' % now
            api.sudo('mkdir -p %s' % backup_folder)
            api.sudo('mv ./src/* %s' % backup_folder)


def update_dev_packes(folder=None, user=None):
    """Update the development packages."""
    config = mls_config()
    folder = folder or config.get('app', {}).get('dir') or err('Folder must be set!')
    user = user or config.get('user') or err('MLS user must be set!')
    with api.settings(sudo_user=user):
        # Update buildout src packages.
        with api.cd(folder):
            api.sudo('./bin/develop up -f')


def run_buildout(folder=None, user=None):
    """Run the buildout."""
    config = mls_config()
    folder = folder or config.get('app', {}).get('dir') or err('Folder must be set!')
    user = user or config.get('user') or err('MLS user must be set!')
    with api.settings(sudo_user=user):
        # Run buildout.
        with api.cd(folder):
            api.sudo('./bin/buildout')


def supervisorctl(command=None, service=None):
    """Control supervisor services."""
    api.sudo('supervisorctl %s %s' % (command, service), warn_only=True)



# -*- coding: utf-8 -*-
"""Deploy and manage Propertyshelf MLS applications using Fabric."""

import bootstrap
import client
import database
import frontend
import roles


bootstrap = bootstrap  # PyFlakes
client = client  # PyFlakes
database = database  # PyFlakes
frontend = frontend  # PyFlakes
roles = roles


__all__ = [
    'bootstrap',
    'client',
    'database',
    'frontend',
    'roles',
]



# -*- coding: utf-8 -*-
"""Manage MLS frontend components like web server, load balancer and cache."""

from fabric import api


@api.task
@api.roles('frontend')
def restart():
    """Restart the frontend components."""
    restart_haproxy()
    restart_varnish()
    restart_nginx()


@api.task
@api.roles('frontend')
def restart_nginx():
    """Restart the NginX web server component."""
    api.sudo('/etc/init.d/nginx restart')


@api.task
@api.roles('frontend')
def restart_varnish():
    """Restart the Varnish caching proxy component."""
    api.sudo('/etc/init.d/varnish restart')


@api.task
@api.roles('frontend')
def restart_haproxy():
    """Restart the HA-Proxy load balancer component."""
    api.sudo('/etc/init.d/haproxy restart')



# -*- coding: utf-8 -*-
"""Manage servers at Rackspace."""

from chef import autoconfigure
from chef.search import Search
from fabric import api
from fabric.colors import red


def create(**params):
    """Create a new server on Rackspace."""
    COMMAND_CREATE = ' '.join([
        'knife rackspace server create',
        '-S %(servername)s',
        '-N %(nodename)s',
        '-f %(flavor)s',
        '-I %(image)s',
        '-r role[rackspace],%(runlist)s',
        '-E %(environment)s',
    ])

    chef_api = autoconfigure()
    query = 'name:%(nodename)s AND chef_environment:%(environment)s' % params
    if len(Search('node', query, api=chef_api)) > 0:
        api.abort(
            'Server "%(nodename)s" already exists in environment '
            '"%(environment)s".' % params)

    api.local(COMMAND_CREATE % params)


def remove(**params):
    """Remove an existing server on Rackspace."""
    COMMAND_LIST = 'knife rackspace server list | grep %s'
    COMMAND_DELETE = 'knife rackspace server delete %s --purge'

    chef_api = autoconfigure()
    query = 'role:%(role)s AND chef_environment:%(environment)s' % params
    nodes = sorted([node.object.name for node in Search('node', query, api=chef_api)])

    print(red('Available nodes:'))
    print(red(', '.join(nodes)))

    client = api.prompt('Which node should be removed?')
    result = api.local(COMMAND_LIST % client, capture=True)
    server_id = result.split()[0]
    api.local(COMMAND_DELETE % server_id)


def next_client_number(**params):
    chef_api = autoconfigure()
    query = 'role:%(rolename)s AND chef_environment:%(environment)s' % params
    nodes = sorted([
        node.object.name for node in Search('node', query, api=chef_api)
    ])
    if len(nodes) < 1:
        return 1
    last = nodes[-1]
    number = last.split('-')[-1]
    try:
        number = int(number)
    except ValueError, e:
        raise e
    else:
        return number + 1



# -*- coding: utf-8 -*-
"""Bootstrap new servers."""

from fabric import api
from mls.fabfile import rackspace
from mls.fabfile.exceptions import missing_env


@api.task
def database(nodename=None, image=None, flavor=None):
    """Bootstrap a new standalone database server."""
    nodename = nodename or api.env.get('nodename_database')
    nodename = nodename or missing_env('nodename_database')

    image = image or api.env.get('os_image')
    image = image or missing_env('os_image')

    flavor = flavor or api.env.get('flavor_database')
    flavor = flavor or missing_env('flavor_database')

    role = api.env.get('role_database')
    role = role or missing_env('role_database')

    runlist = ','.join([
        'role[%s]' % role,
        'recipe[propertyshelf::rackspace_backup]',
    ])

    opts = dict(
        environment='production',
        flavor=flavor,
        image=image,
        nodename=nodename,
        runlist=runlist,
        servername=nodename,
    )
    rackspace.create(**opts)


@api.task
def frontend(nodename=None, image=None, flavor=None):
    """Bootstrap a new standalone frontend server."""
    nodename = nodename or api.env.get('nodename_frontend')
    nodename = nodename or missing_env('nodename_frontend')

    image = image or api.env.get('os_image')
    image = image or missing_env('os_image')

    flavor = flavor or api.env.get('flavor_database')
    flavor = flavor or missing_env('flavor_database')

    role = api.env.get('role_frontend')
    role = role or missing_env('role_frontend')

    runlist = ','.join([
        'role[%s]' % role,
    ])

    opts = dict(
        environment='production',
        flavor=flavor,
        image=image,
        nodename=nodename,
        runlist=runlist,
        servername=nodename,
    )
    rackspace.create(**opts)


@api.task
def worker(nodename=None, image=None, flavor=None):
    """Bootstrap a new standalone application worker."""
    number = rackspace.next_client_number(
        environment='production',
        rolename=api.env.role_worker,
    )

    nodename = nodename or api.env.get('nodename_worker') + '-%02d' % number
    nodename = nodename or missing_env('nodename_worker')

    image = image or api.env.get('os_image')
    image = image or missing_env('os_image')

    flavor = flavor or api.env.get('flavor_worker')
    flavor = flavor or missing_env('flavor_worker')

    role = api.env.get('role_worker')
    role = role or missing_env('role_worker')

    runlist = ','.join([
        'role[%s]' % role,
    ])

    opts = dict(
        environment='production',
        flavor=flavor,
        image=image,
        nodename=nodename,
        runlist=runlist,
        servername=nodename,
    )
    rackspace.create(**opts)


@api.task
def bundle_db_fe_app(nodename=None, image=None, flavor=None):
    """Bootstrap a new MLS bundle: Database, Frontend, App Worker."""
    nodename = nodename or api.env.get('nodename_database')
    nodename = nodename or missing_env('nodename_database')

    image = image or api.env.get('os_image')
    image = image or missing_env('os_image')

    flavor = flavor or api.env.get('flavor_database')
    flavor = flavor or missing_env('flavor_database')

    role_database = api.env.get('role_database')
    role_database = role_database or missing_env('role_database')
    role_frontend = api.env.get('role_frontend')
    role_frontend = role_frontend or missing_env('role_frontend')
    role_worker = api.env.get('role_worker')
    role_worker = role_worker or missing_env('role_worker')

    runlist = ','.join([
        'role[%s]' % role_database,
        'role[%s]' % role_frontend,
        'role[%s]' % role_worker,
        'recipe[propertyshelf::rackspace_backup]',
    ])

    opts = dict(
        environment='production',
        flavor=flavor,
        image=image,
        nodename=nodename,
        runlist=runlist,
        servername=nodename,
    )
    rackspace.create(**opts)


@api.task
def staging(nodename=None, image=None, flavor=None):
    """Bootstrap a staging system."""
    nodename = nodename or api.env.get('nodename_staging')
    nodename = nodename or missing_env('nodename_staging')

    image = image or api.env.get('os_image')
    image = image or missing_env('os_image')

    flavor = flavor or api.env.get('flavor_database')
    flavor = flavor or missing_env('flavor_database')

    role = api.env.get('role_staging')
    role = role or missing_env('role_staging')

    runlist = ','.join([
        'role[%s]' % role,
    ])

    opts = dict(
        environment='staging',
        flavor=flavor,
        image=image,
        nodename=nodename,
        runlist=runlist,
        servername=nodename,
    )
    rackspace.create(**opts)



# -*- coding: utf-8 -*-
"""Manage chef roles."""

from chef import autoconfigure, Role
from fabric import api
from fabric.colors import green, red
from mls.fabfile.exceptions import missing_env


def _check_role(role, roles=None):
    """Check if a given role is available on the chef server."""
    if not roles:
        chef_api = autoconfigure()
        roles = Role.list(api=chef_api)
    return role in roles


def _required_roles():
    """Get a list of required roles."""
    role_database = api.env.get('role_database')
    role_database = role_database or missing_env('role_database')
    role_frontend = api.env.get('role_frontend')
    role_frontend = role_frontend or missing_env('role_frontend')
    role_staging = api.env.get('role_staging')
    role_staging = role_staging or missing_env('role_staging')
    role_worker = api.env.get('role_worker')
    role_worker = role_worker or missing_env('role_worker')

    return {
        'role_database': role_database,
        'role_frontend': role_frontend,
        'role_staging': role_staging,
        'role_worker': role_worker,
    }

@api.task
def check():
    """Check if the required roles are available."""
    chef_api = autoconfigure()
    chef_roles = Role.list(api=chef_api)

    for role in _required_roles().values():
        if _check_role(role, chef_roles):
            print(green('Role %s available.') % role)
        else:
            print(red('Role %s NOT available.') % role)


@api.task
def create_missing():
    """Create missing roles on the chef server."""
    chef_api = autoconfigure()
    required = _required_roles()
    domain = api.env.get('domain', 'example.com')

    # Create the role_database.
    if not _check_role(required.get('role_database')):
        name = required.get('role_database')
        description = 'ZEO Server for %s.' % domain
        run_list = (
            "role[mls_zeoserver]",
        )
        default_attributes = {
            'domain': domain,
        }

        Role.create(
            name,
            api=chef_api,
            description=description,
            run_list=run_list,
            default_attributes=default_attributes,
        )
        print(green('Created role %s') % name)

    # Create the role_frontend.
    if not _check_role(required.get('role_frontend')):
        name = required.get('role_frontend')
        description = 'Frontend Server for %s.' % domain
        run_list = (
            "role[mls_loadbalancer]",
        )
        default_attributes = {
            'domain': domain,
            'haproxy': {
                'app_server_role': required.get('role_worker'),
            },
            'mls': {
                'domain': '.'.join(['mls', domain]),
            },
        }

        Role.create(
            name,
            api=chef_api,
            description=description,
            run_list=run_list,
            default_attributes=default_attributes,
        )
        print(green('Created role %s') % name)

    # Create the role_worker.
    if not _check_role(required.get('role_worker')):
        name = required.get('role_worker')
        description = 'Application Worker for %s.' % domain
        run_list = (
            "role[mls_application]",
        )
        default_attributes = {
            'domain': domain,
            'mls': {
                'customizations': api.env.get('mls_customizations', []),
                'policy': {
                    'enabled': api.env.get('mls_policy_enabled') and 'true' or 'false',
                    'package': api.env.get('mls_policy_package', ''),
                    'package_url': api.env.get('mls_policy_package_url', ''),
                },
                'zeo_role': required.get('role_database'),
            },
        }

        Role.create(
            name,
            api=chef_api,
            description=description,
            run_list=run_list,
            default_attributes=default_attributes,
        )
        print(green('Created role %s') % name)

    # Create the role_staging.
    if not _check_role(required.get('role_staging')):
        name = required.get('role_staging')
        description = 'Staging system for %s.' % domain
        run_list = (
            "role[%s]" % required.get('role_database'),
            "role[%s]" % required.get('role_worker'),
            "role[%s]" % required.get('role_frontend'),
        )
        default_attributes = {
            'domain': domain,
            'haproxy': {
                'app_server_role': name,
            },
            'mls': {
                'domain': '.'.join(['staging', domain]),
                'zeo_role': name,
            },
        }

        Role.create(
            name,
            api=chef_api,
            description=description,
            run_list=run_list,
            default_attributes=default_attributes,
        )
        print(green('Created role %s') % name)



# -*- coding: utf-8 -*-
"""Manage MLS application client components."""

from fabric import api
from mls.fabfile import rackspace, utils
from mls.fabfile.exceptions import missing_env


@api.task
def remove():
    """Remove an existing MLS application client."""
    role = api.env.get('role_worker')
    role = role or missing_env('role_worker')
    opts = dict(
        environment='production',
        role=role,
    )
    rackspace.remove(**opts)


@api.task
@api.roles('worker')
def update():
    """Update the client packages."""
    utils.supervisorctl(command='stop', service='application')
    utils.backup_dev_packages()
    utils.run_buildout()
    utils.supervisorctl(command='start', service='application')


@api.task
@api.roles('worker')
def restart():
    """Restart the application client component."""
    utils.supervisorctl(command='restart', service='application')



# -*- coding: utf-8 -*-
"""Environment definitions."""

from chef.fabric import chef_roledefs
from fabric import api
from mls.fabfile.exceptions import missing_env

__all__ = [
    'development',
    'production',
    'staging',
]


def update_roledefs(environment):
    """Update Fabric's role definitions with our chef based roles.

    :param environment: The chef environment to search for.
    :type environment: string
    """
    roledefs = chef_roledefs(
        hostname_attr=['ipaddress'],
        environment=environment,
    )
    role_database = api.env.get('role_database', None)
    if role_database:
        api.env.roledefs['database'] = roledefs.get(role_database)

    role_frontend = api.env.get('role_frontend')
    if role_frontend:
        api.env.roledefs['frontend'] = roledefs.get(role_frontend)

    role_staging = api.env.get('role_staging')
    if role_staging:
        api.env.roledefs['staging'] = roledefs.get(role_staging)

    role_worker = api.env.get('role_worker')
    if role_worker:
        api.env.roledefs['worker'] = roledefs.get(role_worker)


@api.task
def development():
    """Work locally with vagrant."""
    # Change the default user to 'vagrant'.
    result = api.local('vagrant ssh-config | grep IdentityFile', capture=True)
    api.env.key_filename = result.replace('"', '').split()[1]
    api.env.user = 'vagrant'

    # Connect to the port-forwarded ssh.
    api.env.hosts = ['127.0.0.1:2222']


    nodename = api.env.get('nodename_development')
    nodename = nodename or api.env.get('nodename_staging')
    nodename = nodename or missing_env('nodename_staging')

    api.env.hostname = 'vagrant-%(name)s-%(user)s' % {
        'name': nodename,
        'user': api.env.local_user,
    }

    # Set role definitions for vagrant.
    api.env.roledefs = {
        'worker': ['127.0.0.1:2222', ],
        'frontend': ['127.0.0.1:2222', ],
        'database': ['127.0.0.1:2222', ],
    }


@api.task
def staging():
    """Work with the staging environment."""
    update_roledefs('staging')


@api.task
def production():
    """Work with the production environment."""
    update_roledefs('production')



# -*- coding: utf-8 -*-
"""Installer for mls.fabfile."""

from setuptools import setup, find_packages

version = '0.2'

long_description = (
    open('README.rst').read()
    + '\n' +
    'Contributors\n'
    '============\n'
    + '\n' +
    open('CONTRIBUTORS.rst').read()
    + '\n' +
    open('CHANGES.rst').read()
    + '\n')

setup(
    name='mls.fabfile',
    version=version,
    description='Deploy and manage Propertyshelf MLS applications using Fabric.',
    long_description=long_description,
    # Get more strings from
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Programming Language :: Python',
    ],
    keywords='fabric python mls deployment',
    author='Propertyshelf, Inc.',
    author_email='development@propertyshelf.com',
    url='https://github.com/propertyshelf/mls.fabfile',
    download_url='http://pypi.python.org/pypi/mls.fabfile',
    license='BSD',
    packages=find_packages('src', exclude=['ez_setup']),
    package_dir = {'': 'src'},
    namespace_packages=['mls'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        # -*- Extra requirements: -*-
        'Fabric',
        'PyChef',
    ],
    entry_points="""
    # -*- Entry points: -*-
    """,
)



# -*- coding: utf-8 -*-
"""Manage MLS database components."""

import os
from fabric import api
from fabric.contrib.console import confirm
from fabric.contrib.files import exists
from mls.fabfile import utils
from mls.fabfile.exceptions import err


@api.task
@api.roles('database')
def restart():
    """Restart the database component."""
    utils.supervisorctl(command='restart', service='zeoserver')


@api.task
@api.roles('database')
def download_data():
    """Download the database files from the server."""
    download_zodb()
    download_blobs()


@api.task
@api.roles('database')
def download_zodb():
    """Download ZODB part of Zope's data from the server."""
    config = utils.mls_config()
    folder = config.get('zeo', {}).get('dir') or err('Folder must be set!')
    user = config.get('user') or err('MLS user must be set!')

    if not confirm('This will overwrite your local Data.fs. Are you sure you '
                   'want to continue?'):
        api.abort('ZODB download cancelled.')

    api.local('mkdir -p var/filestorage')

    # Backup current Data.fs.
    if os.path.exists('var/filestorage/Data.fs'):
        api.local('mv var/filestorage/Data.fs var/filestorage/Data.fs.bak')

    with api.settings(sudo_user=user):

        # Remove temporary Data.fs file from previous downloads.
        if exists('/tmp/Data.fs', use_sudo=True):
            api.sudo('rm -rf /tmp/Data.fs')

        # Downlaod Data.fs from server.
        with api.cd(folder):
            api.sudo('rsync -a var/filestorage/Data.fs /tmp/Data.fs')
            api.get('/tmp/Data.fs', 'var/filestorage/Data.fs')


@api.task
@api.roles('database')
def download_blobs():
    """Download blob part of Zope's data from the server."""
    config = utils.mls_config()
    folder = config.get('zeo', {}).get('dir') or err('Folder must be set!')
    user = config.get('user') or err('MLS user must be set!')

    if not confirm('This will overwrite your local blob files. Are you sure '
                   'you want to continue?'):
        api.abort('Blob download cancelled.')

    # Remove local blob files backup.
    if os.path.exists('var/blobstorage_bak'):
        api.local('rm -rf var/blobstorage_bak')

    # Backup current blob files.
    if os.path.exists('var/blobstorage'):
        api.local('mv var/blobstorage var/blobstorage_bak')

    with api.settings(sudo_user=user):

        # Remove temporary blob files from previous downloads.
        if exists('/tmp/blobstorage', use_sudo=True):
            api.sudo('rm -rf /tmp/blobstorage')

        if exists('/tmp/blobstorage.tgz', use_sudo=True):
            api.sudo('rm -rf /tmp/blobstorage.tgz')

        # Download blob files from server.
        with api.cd(folder):
            api.sudo('rsync -a ./var/blobstorage /tmp/')

        with api.cd('/tmp'):
            api.sudo('tar czf blobstorage.tgz blobstorage')
            api.get('/tmp/blobstorage.tgz', './var/blobstorage.tgz')

        with api.lcd('var'):
            api.local('tar xzf blobstorage.tgz')


@api.task
@api.roles('database')
def upload_data():
    """Upload the database files to the server."""
    upload_zodb(start_when_done=False)
    upload_blob()


@api.task
@api.roles('database')
def upload_zodb(start_when_done=True):
    """Upload ZODB part of Zope's data to the server."""
    config = utils.mls_config()
    folder = config.get('zeo', {}).get('dir') or err('Folder must be set!')
    user = config.get('user') or err('MLS user must be set!')

    if not confirm('This will overwrite your remote Data.fs. Are you sure you '
                   'want to continue?', default=False):
        api.abort('ZODB upload cancelled.')

    api.sudo('mkdir -p /tmp/upload', user=user)

    api.put('var/filestorage/Data.fs', '/tmp/upload/Data.fs', use_sudo=True)
    api.sudo('chown %s /tmp/upload/Data.fs' % user)

    utils.supervisorctl(command='stop', service='zeoserver')
    with api.settings(sudo_user=user):
        with api.cd(folder):
            # Backup current Data.fs.
            if exists('var/filestorage/Data.fs'):
                api.sudo('mv var/filestorage/Data.fs var/filestorage/Data.fs.bak')
            api.sudo('mv /tmp/upload/Data.fs var/filestorage/Data.fs')

    if start_when_done:
        utils.supervisorctl(command='start', service='zeoserver')


@api.task
@api.roles('database')
def upload_blob(start_when_done=True):
    """Upload blob part of Zope's data to the server."""
    config = utils.mls_config()
    folder = config.get('zeo', {}).get('dir') or err('Folder must be set!')
    user = config.get('user') or err('MLS user must be set!')

    if not confirm('This will overwrite your remote blob files. Are you sure '
                   'you want to continue?', default=False):
        api.abort('Blob upload cancelled.')

    api.sudo('mkdir -p /tmp/upload', user=user)

    with api.lcd('var'):
        api.local('tar czf blobstorage_upload.tgz blobstorage')
        api.put('blobstorage_upload.tgz', '/tmp/upload/blobstorage.tgz', use_sudo=True)

    api.sudo('chown %s /tmp/upload/blobstorage.tgz' % user)
    with api.cd('/tmp/upload'):
        api.sudo('tar xzf blobstorage.tgz', user=user)

    utils.supervisorctl(command='stop', service='zeoserver')
    with api.settings(sudo_user=user):
        with api.cd(folder):
            # Backup current blob files.
            if exists('var/blobstorage'):
                api.sudo('mv var/blobstorage var/blobstorage_bak')
            api.sudo('mv /tmp/upload/blobstorage var')

    if start_when_done:
        utils.supervisorctl(command='start', service='zeoserver')


@api.task
@api.roles('database')
def backup():
    """Perform a backup of Zope's data on the server."""
    raise NotImplementedError


@api.task
@api.roles('database')
def restore():
    """Restore an existing backup of Zope's data on the server."""
    raise NotImplementedError
